unit CuBlasMtxVec;

interface

uses  SysUtils, CuBlas, Cublas_API, Math, Classes, UCudaRT_V2;

type
  TVec = Class
  private
    function GetValue(index: integer): Double;
    procedure SetValue(index: integer; const Value: Double);
  protected
  FLength: integer;
  Values: Array of Double;
  public
  constructor Create;
  destructor Destroy; override;
  procedure Size(n: integer);
  procedure SetZero;
  procedure Assign(Vec: TVec);
  function Add(Vec: TVec): TVec;
  function Sub(Vec: TVec): TVec;
  function Dot(Vec: TVec): Double;
  procedure Resize(n: integer);
  function NormL2: Double;
  function Scale(value: Double): TVec;
  function GetSubVec(index: integer): TVec;
  procedure CopyVecs(Vec1, Vec2: TVec);
  procedure CopyVal(Vec: TVec; VecIndex, Index, Len: integer);
  property length: integer read Flength;
  property Value[index: integer]: Double read GetValue write SetValue; default;
  end;

  TMtx = Class
  private
    function GetValue(col, row: integer): Double;
    procedure SetValue(col, row: integer; const Value: Double);
  protected
  FLength: integer;
  Frank: integer;
  Values: Array of Double;
  public  
  constructor Create;
  destructor Destroy; override;
  procedure Size(col, row: integer);
  procedure Assign(Mtx: TMtx);
  function Add(Mtx: TMtx): TMtx;
  function Sub(Mtx: TMtx): TMtx;
  function NormL2: Double;
  function Prod(Vec: TVec): TVec;
  function TriSolver(Vec: TVec): TVec;
  procedure Scale(Value: Double);
  procedure SetCol(Vec: TVec; Col: integer);
  function GetCol(Col: integer): TVec;
  function GetSubMtx(Row, Col: integer): TMtx;
  function CG(B: TVec; var X: TVec): integer;
  function BiCGStab(B: TVec; var X: TVec): integer;
  function GMRES(B: TVec; var X: TVec): integer;
  property Value[col, row: integer]: Double read GetValue write SetValue; default;
  end;

procedure DecomposeLU(M, N, lda: integer; A: TMtx; P: array of integer; epsilon: double);

implementation

var
  SelfValues, VecValues, VecValues1,
  xx, r, r_tld, p, p_hat, q, s, sdev, s_hat, t, v,
  y, vi, vk, w, hh: PDoubleVector;
  Ipiv: PIntegerVector;
  CulaSelfValues, CulaVecValues, CulaVecValues1: array of Double;
  IPIVec: array of Integer;
  statuscublas: cublasStatus_t;
  cudaError: cudaError_t;
  errorr: cardinal;
  cuda_error: byte;

//***************************************************************************************
//void DecomposeLU( int M, int N, int lda , float* A,   
//                    int* permute, float epsilon, InfoStat& stat)
//
// M         :    Num of rows of A
// N         :    Num of column of A
// A         :    Float Matrix of size M*N
//                on the output contains the result of the LU decomposition
//                The diagonal elements for L are not stored in A ( assuming they are all 1)
//lda        :    Leading dim of A lda < std::max(1,M)
//P          :    Permutation vector of size M
//epsilon    :    Epsilon (used to test for singularities)
//stat       :    return status

// **************************************************************************************
procedure DecomposeLU(M, N, lda: integer; A: TMtx; P: array of integer; epsilon: double);
var
     cuStat: cublasStatus;
     minDim, k, pivotRow, kp1: integer;
     valcheck: double;
     //Preconditions
begin
  statuscublas:= cublasAlloc(N, sizeof(double), @SelfValues);
  statuscublas:= cublasSetVector(N, sizeof(double),@A.Values[0],1,@SelfValues[0],1);
  if (  (M<=0) or (N<=0) or (lda < max(1,M)) ) then
     begin
          //stat._info = -1;
          if (M<=0) then
              //stat._str = 'M<=0';
          if (N<=0) then
              //stat._str = 'M<=0';
          if (lda < max(1,M)) then
              //stat._str = "lda &lt; std::max(1,M)";
          exit;
     end;
     minDim := min( M, N );
  for k:=0 to minDim-1 do
     begin
          pivotRow := k-1+cublasIsamax(M-k, @SelfValues[k + k*lda], 1); // row relative to the current submatrix
          kp1 := k+1;
          P[k] := pivotRow;
          if (pivotRow <> k) then
          begin
               cublasDswap(N, SelfValues[pivotRow], lda, SelfValues[k], lda);
          end;
          cublasGetVector(1, sizeof(double), @SelfValues[k+ k*lda], 1, @valcheck, 1);
          if (abs(valcheck) < epsilon) then
          begin
               //stat._info =k+1;
               //stat._str = " Matrix is Singular ";
               exit;
          end;
          if (kp1 < M) then
          begin
              cublasDscal(M-kp1, 1.0/valcheck, SelfValues[kp1+ k*lda], 1);
          end;
         if (  kp1 < minDim  ) then
         begin
              cublasDger(M-kp1, N-kp1, -1.0, SelfValues[kp1+ k*lda], 1, SelfValues[k+ kp1*lda], lda, SelfValues[kp1*lda+kp1], lda);
         end;
     end;
     //CHECK_CUBLAS('decomposeLU pb');
  statuscublas:= cublasGetVector(N, sizeof(double),@SelfValues[0],1,@A.Values[0],1);
  statuscublas:= cublasFree(@SelfValues[0]);
end;


//***************************************************************************************
//void DecomposeBlockedLU (	int M, int N,int lda,
//							float *A,
//							int* P, int blockSize,float epsilon, InfoStat &amp;stat )
//
// M 			:  Num of rows of A
// N 			:  Num of column of A
// A 			:  Float Matrix of size M*N
//					 on the output contains the result of the LU decomposition
//					 The diagonal elements for L are not stored in A ( assuming they are all 1)
//lda			:	 Leading dim of A lda &lt; std::max(1,M) //P 			:  	Permutation vector of size M //blockSize		:	Size of the submatrices //					if blockSize&gt;=M || blockSize==1 unblocked decomposition is called
//epsilon	:  Epsilon (used to test for singularities)
//stat 	 	:  return status
// **************************************************************************************
procedure DecomposeBlockedLU ( M, N, lda: integer; A: TMtx; P: array of integer; blockSize: integer; epsilon: single);
var
     cuStat: cublasStatus;
     minSize, realBlockSize, step, i, pp: integer;
     valcheck: single;
     //Preconditions
begin
  statuscublas:= cublasAlloc(N, sizeof(double), @SelfValues);
  statuscublas:= cublasSetVector(N, sizeof(double),@A.Values[0],1,@SelfValues[0],1);

  if (  (M<=0) or (N<=0) or (lda < max(1,M)) ) then
     begin
          //stat._info = -1;
     if (M<=0) then
             //stat._str = 'M<=0';
     if (N<=0) then
              //stat._str = 'M<=0';
     if (lda < max(1,M)) then
              //stat._str = "lda &lt; std::max(1,M)";
     exit;
     end;
  minSize := min(M,N);
  if ( (blockSize > minSize) or (blockSize = 1)) then
	begin
		//straight LU decomposition
		DecomposeLU( M, N, lda, A, P, epsilon);
	end
	else
	begin
		//blocked decomposition
    i:= 0;
		while  (i < minSize-1) do   //i+=blockSize
		begin
      step := i*blockSize;
			realBlockSize:= min(minSize - i, blockSize);

			//decompose the current rectangular block
		 DecomposeLU( M-i, realBlockSize, lda, @SelfValues[i+i*lda], P[i], epsilon);

			//adjust pivot infos
			//Todo : write a kernel for that
			for pp:= i to min( M, i+realBlockSize)-1 do
			begin
					P[pp] := P[pp+i];
					if (P[pp] <> pp) then
					begin
						// Apply interchanges to columns 0:i.
						cublasDswap(i, SelfValues[pp] , lda, SelfValues[P[pp]], lda);
						// Apply interchanges to columns i+blockSize:N.
						cublasDswap(N-i-realBlockSize, SelfValues[pp+(i+realBlockSize)*lda], lda, SelfValues[P[pp]+(i+realBlockSize)*lda], lda);
					end;
    i:= i+1;
		end;

			// Compute block row of U.
			cublasDtrsm( 'l','l','n','u', realBlockSize, N-i-realBlockSize, 1.0,
						 SelfValues[i+i*lda], lda, SelfValues[i + (i+realBlockSize)*lda], lda);
			//CHECK_CUBLAS("decomposeBlockedLU cublasStrsm");

			if (i+realBlockSize < M) then
			begin
				 cublasDgemm('n','n',  M-i-realBlockSize, N-i-realBlockSize, realBlockSize,
							 -1.0,
							 SelfValues[i+realBlockSize+i*lda],lda,
							 SelfValues[i+(realBlockSize+i)*lda],lda,
							 1.0,
							 SelfValues[i+realBlockSize+(realBlockSize+i)*lda],lda );
				 //CHECK_CUBLAS("decomposeBlockedLU cublasSgemm");
			end;
		end;
	end;
  statuscublas:= cublasGetVector(N, sizeof(double),@SelfValues[0],1,@A.Values[0],1);
  statuscublas:= cublasFree(@SelfValues[0]);
end;


{ TculaVec }

constructor TVec.Create;
begin
  inherited Create;
  SetLength(values, 0);
end;

procedure TVec.Size(n: integer);
begin
  Flength:= n;
  SetLength(values, n);
end;

procedure TMtx.Assign(Mtx: TMtx);
begin
  Self.Values:= Copy(Mtx.Values, 0, FLength);
end;


function TVec.GetValue(index: integer): Double;
begin
  result:= Values[index];
end;

procedure TVec.SetValue(index: integer; const Value: Double);
begin
  Values[index]:= Value;
end;

procedure TVec.CopyVecs(Vec1, Vec2: TVec);
begin
  Move(Vec1.Values,Values[0],Vec1.FLength*sizeof(Values[0]));
  Move(Vec2.Values,Values[Vec1.FLength],Vec2.FLength*sizeof(Values[0]));
end;

procedure TVec.CopyVal(Vec: TVec; VecIndex, Index, Len: integer);
begin
  Values:= Copy(Vec.Values, Index, Len);
end;

function TVec.Add(Vec: TVec): TVec;
begin
  Result:= TVec(Self);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Flength, sizeof(double), @VecValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Flength, sizeof(double),@Vec.Values[0],1,@VecValues[0],1);

  cublasDaxpy(Flength,1.0,VecValues[0],1,SelfValues[0],1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@SelfValues[0]);

end;

function TVec.Sub(Vec: TVec): TVec;
begin
  Result:= TVec(Self);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Flength, sizeof(double), @VecValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Flength, sizeof(double),@Vec.Values[0],1,@VecValues[0],1);

  cublasDaxpy(Flength,-1.0,VecValues[0],1,SelfValues[0],1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@SelfValues[0]);

end;

function TVec.GetSubVec(index: integer): TVec;
var i: integer;
begin
  Result:= TVec.Create;
  Result.Size(index);
  for i:= 0 to index-1 do
    Result[i]:= Self[i];
end;

function TVec.NormL2: Double;
begin

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);

  result:= cublasDnrm2(Flength, SelfValues[0], 1);

  statuscublas:= cublasFree(@SelfValues[0]);

end;

destructor TVec.Destroy;
begin
  Finalize(Values,0);
  inherited;
end;



function TVec.Dot(Vec: TVec): Double;
begin

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Flength, sizeof(double), @VecValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Flength, sizeof(double),@Vec.Values[0],1,@VecValues[0],1);

  result:= cublasDdot(Flength,VecValues[0],1,SelfValues[0],1);

  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@SelfValues[0]);

end;

function TVec.Scale(value: Double): TVec;
begin
  result:= TVec(Self);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);

  cublasDscal(Self.FLength, value, SelfValues[0], 1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@SelfValues[0]);
end;

procedure TVec.SetZero;
var
  value: double;
begin
  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);

  value:= 0.0;
  cublasDscal(Self.FLength, value, SelfValues[0], 1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Self.Values[0],1);
end;

procedure TVec.Assign(Vec: TVec);
begin
  Self.Values:= Copy(Vec.Values, 0, FLength);
end;

procedure TVec.Resize(n: integer);
begin
  Size(n);
end;

{ TculaMtx }

function TMtx.Add(Mtx: TMtx): TMtx;
begin
  Result:= TMtx(Self);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Flength, sizeof(Mtx.Values[0]), @VecValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Flength, sizeof(Mtx.Values[0]),@Mtx.Values[0],1,@VecValues[0],1);

  cublasDaxpy(Flength,1.0,VecValues[0],1,SelfValues[0],1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@SelfValues[0]);

end;

constructor TMtx.Create;
begin
  SetLength(values, 0);
end;

procedure TMtx.Size(col, row: integer);
begin
  Frank:= col;
  Flength:= col*row;
  SetLength(Values, Flength);
end;

destructor TMtx.Destroy;
begin
  Finalize(Values,0);
  inherited;
end;

function TMtx.GetValue(col, row: integer): Double;
begin
  result:= Values[IDX2(col, row, Frank)];
end;

procedure TMtx.SetCol(Vec: TVec; Col: integer);
var i: integer;
begin
  Move(Vec.Values[0], Values[Vec.Flength*Col], Vec.Flength*sizeof(Values[0]));
end;

function TMtx.GetCol(Col: integer): TVec;
begin
  Result:= TVec.Create;
  Result.Size(Frank);
  Move(Values[Frank*Col], Result.Values[0], Frank*sizeof(Values[0]));
end;

function TMtx.GetSubMtx(Row, Col: integer): TMtx;
var i, j: integer;
begin
  Result:= TMtx.Create;
  Result.Size(row, col);
    for i:= 0 to row-1 do
      for j:= 0 to col-1 do
        Result[i,j]:= Self[i,j];
end;

function TMtx.NormL2: Double;
begin

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);

  result:= cublasDnrm2(Flength, SelfValues[0], 1);

  statuscublas:= cublasFree(@SelfValues[0]);

end;

function TMtx.Prod(Vec: TVec): TVec;
begin
  Result:= TVec.Create;
  Result.Size(Vec.FLength);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Vec.Flength, sizeof(double), @VecValues);
  statuscublas:= cublasAlloc(Result.Flength, sizeof(double), @VecValues1);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Vec.Flength, sizeof(double),@Vec.Values[0],1,@VecValues[0],1);
  statuscublas:= cublasSetVector(Result.Flength, sizeof(Result.Values[0]),@Result.Values[0],1,@VecValues1[0],1);

  cublasDgemv('n',Vec.Flength,Vec.Flength,1.0,SelfValues[0],Vec.FLength,VecValues[0],1,0.0,VecValues1[0],1);

  statuscublas:= cublasGetVector(Vec.Flength, sizeof(double),@VecValues1[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@SelfValues[0]);
  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@VecValues1[0]);

end;

procedure TMtx.Scale(value: Double);
begin

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);

  cublasDscal(Self.FLength, value, SelfValues[0], 1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Self.Values[0],1);

end;

procedure TMtx.SetValue(col, row: integer; const Value: Double);
begin
  Values[IDX2(col, row, Frank)]:= Value;
end;


function TMtx.Sub(Mtx: TMtx): TMtx;
begin
  Result:= TMtx(Self);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Flength, sizeof(Mtx.Values[0]), @VecValues);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Flength, sizeof(Mtx.Values[0]),@Mtx.Values[0],1,@VecValues[0],1);

  cublasDaxpy(Flength,-1.0,VecValues[0],1,SelfValues[0],1);

  statuscublas:= cublasGetVector(Flength, sizeof(double),@SelfValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@VecValues[0]);
  statuscublas:= cublasFree(@SelfValues[0]);

end;

function TMtx.TriSolver(Vec: TVec): TVec;
begin
  Result:= TVec.Create;
  Result.Size(Vec.FLength);

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(Vec.Flength, sizeof(double), @VecValues);


  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(Vec.Flength, sizeof(double),@Vec.Values[0],1,@VecValues[0],1);


  cublasDtrsv('u','n','n',Vec.Flength,SelfValues[0],Vec.Flength,VecValues[0],1);

  statuscublas:= cublasGetVector(Vec.Flength, sizeof(double),@VecValues[0],1,@Result.Values[0],1);

  statuscublas:= cublasFree(@SelfValues[0]);
  statuscublas:= cublasFree(@VecValues[0]);


end;

const MAX_IT = 20;
      TOL = 1e-7;

function TMtx.CG(B: TVec; var X: TVec): integer;
label start, finish;
var i, N: integer;
    nrmr0, rhop, rho, beta, temp, alpha, nrmr: double;
begin
  N:= B.FLength;

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasAlloc(N, sizeof(double), @r);
  statuscublas:= cublasAlloc(N, sizeof(double), @p);
  statuscublas:= cublasAlloc(N, sizeof(double), @q);
  statuscublas:= cublasAlloc(N, sizeof(double), @xx);

  statuscublas:= cublasSetVector(N, sizeof(X.Values[0]),@X.Values[0],1,@xx[0],1);

  start:
  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(N, sizeof(double),@B.Values[0],1,@r[0],1);
  
  // r := f - A x0 ( using initial guess in x )
  cublasDgemv('n', N, N, -1.0, SelfValues[0], N, xx[0], 1, 1.0, r[0], 1);
  nrmr0 := cublasDnrm2 ( n , r[0] , 1) ;
  rho:= nrmr0;
  // repeat until convergence ( based on max . i t . and relative residual )
  for i:=0 to MAX_IT do
   begin
   // Mz = r ( sparse lower and upper triangular solves )
   // rho := r^{T} z
   rhop:= rho ;
   rho := cublasDdot ( n , r[0] , 1 , r[0] , 1) ;
   if ( i = 0) then
     begin
     // p := z
     cublasDcopy ( n , r[0] , 1 , p[0] , 1) ;
     end
     else
     begin
     // beta := rho_{i} / rho_{i-1}
     beta:= rho / rhop ;
     // p := z + beta p
     cublasDaxpy ( n , beta , p[0] , 1 , r[0] , 1) ;
     cublasDcopy ( n , r[0] , 1 , p[0] , 1) ;
     end;
   // q := A*p ( sparse matrix-vector multiplication )
   cublasDgemv('n', N, N, 1.0, SelfValues[0], N, p[0], 1, 0.0, q[0], 1);
   // alpha := rho_{i} / (p^{T} q )
   temp := cublasDdot ( n , p[0] , 1 , q[0] , 1) ;
   if abs(temp) < TOL then goto finish;
   alpha:= rho / temp ;
   // x := x + alpha p
   cublasDaxpy ( n , alpha , p[0] , 1 , xx[0] , 1) ;
   // r := r - alpha q
   cublasDaxpy ( n, -alpha , q[0] , 1 , r[0] , 1) ;
   // check for convergence
   nrmr:= cublasDnrm2 ( n , r[0] , 1) ;
   if ( (nrmr / nrmr0) < tol ) then
      break ;
  end;
  if ((nrmr / nrmr0) > tol) then goto start;

  finish:
  statuscublas:= cublasGetVector(N, sizeof(X.Values[0]),@xx[0],1,@X.Values[0],1);
  statuscublas:= cublasFree(@SelfValues[0]);
  statuscublas:= cublasFree(@r[0]);
  statuscublas:= cublasFree(@p[0]);
  statuscublas:= cublasFree(@q[0]);
  statuscublas:= cublasFree(@xx[0]);

  result:= 0;
end;

function TMtx.BiCGStab(B: TVec; var X: TVec): integer;
label Start, finish;
var i: integer;
    iter, flag, M, N: integer;
    bnrm2, snrm2, error, alpha, beta, omega, rho, rho_1, resid: double;
begin
  N:= B.FLength;

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);

  statuscublas:= cublasAlloc(N, sizeof(double), @r);
  statuscublas:= cublasAlloc(N, sizeof(double), @s);
  statuscublas:= cublasAlloc(N, sizeof(double), @t);
  statuscublas:= cublasAlloc(N, sizeof(double), @r_tld);
  statuscublas:= cublasAlloc(N, sizeof(double), @p);
  statuscublas:= cublasAlloc(N, sizeof(double), @p_hat);
  statuscublas:= cublasAlloc(N, sizeof(double), @s_hat);
  statuscublas:= cublasAlloc(N, sizeof(double), @v);
  statuscublas:= cublasAlloc(N, sizeof(double), @xx);

  statuscublas:= cublasSetVector(N, sizeof(X.Values[0]),@X.Values[0],1,@xx[0],1);

  start:
  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(N, sizeof(double),@B.Values[0],1,@r[0],1);

  bnrm2 := cublasDnrm2(N, r[0], 1);

  //* r = r - A*x; r is now residual */
  cublasDgemv('n', N, N, -1.0, SelfValues[0], N, xx[0], 1, 1.0, r[0], 1);

  error := cublasDnrm2 (N, r[0], 1) / bnrm2;   //* norm_r = norm(b) */
  if ( error < TOL ) then goto finish;   //* x is close enough already */

  omega := 1.0;

  //* r_tld = r */
  cublasDCopy(N, r[0], 1, r_tld[0], 1);

  for iter := 0 to MAX_IT-1 do
  begin

    rho := cublasDdot(N, r_tld[0], 1, r[0], 1);  //* rho = r_tld'*r */

    if rho = 0.0 then goto finish;

    if ( iter > 0 ) then
	  begin
	  beta := (rho/rho_1) * (alpha/omega);
	  cublasDaxpy (N, -omega, v[0], 1, p[0], 1);
	  cublasDaxpy (N, 1.0/beta, r[0], 1, p[0], 1);
	  cublasDscal (N, beta, p[0], 1);       //* p = r + beta*( p - omega*v ) */
	  end
    else
	   cublasDcopy(N, r[0], 1, p[0], 1); //* p = r */

    //* p_hat = M \ p    --> here assume M = I, the identity matrix */
    cublasDcopy(N, p[0], 1, p_hat[0], 1);   //* p_hat = p */

    //* v = A*p_hat */
    cublasDgemv('n', N, N, 1.0, SelfValues[0], N, p_hat[0], 1, 0.0, v[0], 1);

    alpha := rho / cublasDdot(N, r_tld[0], 1, v[0], 1);  //* alph = rho / ( r_tld'*v ) */

    cublasDcopy(N, r[0], 1, s[0], 1); //* s = r */

    cublasDaxpy(N, -alpha, v[0], 1, s[0], 1);
    snrm2 := cublasDnrm2( N, s[0], 1);

    if ( snrm2 < TOL ) then
	  begin
	  cublasDaxpy (N, alpha, p_hat[0], 1, s[0], 1);
	  resid := snrm2 / bnrm2;
	  break;
	  end;

    //* s_hat = M \ s    --> here assume M = I, the identity matrix */

    //* s_hat = s */
    cublasDcopy(N, s[0], 1, s_hat[0], 1);

    //* t = A*s_hat */
    cublasDgemv('n', N, N, 1.0, SelfValues[0], N, s_hat[0], 1, 0.0, t[0], 1);

    //* omega = ( t'*s) / ( t'*t ) */
    omega := cublasDdot(N, t[0], 1, s[0], 1)/cublasDdot(N, t[0], 1, t[0], 1);

    cublasDaxpy(N, alpha, p_hat[0], 1, xx[0], 1);
    cublasDaxpy(N, omega, s_hat[0], 1, xx[0], 1);  //*  x = x + alph*p_hat + omega*s_hat */

    cublasDaxpy(N, -omega, t[0], 1, s[0], 1);

    //* r = s - omega*t */
    cublasDcopy(N, s[0], 1, r[0], 1);

    error := cublasDnrm2( N, r[0], 1) / bnrm2;
    if ( snrm2 <= TOL ) then goto finish;
    if ( omega = 0.0 ) then goto finish;
    rho_1 := rho;
  end;

  if ( error <= TOL ) then
    flag := 0
  else if ( omega = 0.0 ) then
    flag := -2
  else if ( rho = 0.0 ) then
    flag := -1
  else
    flag := 1;

  if (error > TOL ) then goto start;

  {if ( flag <> 0 ) then

  ShowMessage(Format('BiCGStab produced answer with resid %e in %d iterations ',[resid, iter]))
  else
    ShowMessage(Format('BiCGStab produced error %d after %d iterations ',[flag, iter])); }

  finish:
  statuscublas:= cublasGetVector(N, sizeof(X.Values[0]),@xx[0],1,@X.Values[0],1);
  statuscublas:= cublasFree(@SelfValues[0]);
  statuscublas:= cublasFree(@r[0]);
  statuscublas:= cublasFree(@s[0]);
  statuscublas:= cublasFree(@t[0]);
  statuscublas:= cublasFree(@r_tld[0]);
  statuscublas:= cublasFree(@p[0]);
  statuscublas:= cublasFree(@p_hat[0]);
  statuscublas:= cublasFree(@s_hat[0]);
  statuscublas:= cublasFree(@v[0]);
  statuscublas:= cublasFree(@xx[0]);

  result:= 0;
end;

procedure rotmat( a, b: double; var sn1, cs1: double);
var temp: double;
begin
 //
 // Compute the Givens rotation matrix parameters for a and b.
 //
  if ( b = 0.0 ) then begin
      cs1 := 1.0;
      sn1 := 0.0;   end
  else if ( abs(b) > abs(a) ) then begin
      temp := a / b;
      sn1 := 1.0 / sqrt( 1.0 + Power(temp,2) );
      cs1 := temp * sn1; end
   else begin
      temp := b / a;
      cs1 := 1.0 / sqrt( 1.0 + Power(temp,2) );
      sn1 := temp * cs1;
   end;
end;

function TMtx.GMRES(B: TVec; var X: TVec): integer;
label
   finish;
var
  N, M, iter, j, i, k, kp1, col, row: integer;
  bnrm2, error, rnorm2, sni, csi, temp: Double;
  H: array of Double;
  cs, sn, s, hhh, vvv: array of Double;

begin
  N:= B.FLength;
  M:= N;
  setlength(H,(m+1)*m);
  setlength(s,m+1);
  setlength(cs,m);
  setlength(sn,m);
  s[0]:= 1.0;

  statuscublas:= cublasAlloc(Flength, sizeof(double), @SelfValues);
  statuscublas:= cublasAlloc(N, sizeof(double), @r);
  statuscublas:= cublasAlloc(N*(m+1), sizeof(double), @V);
  statuscublas:= cublasAlloc(N, sizeof(double), @vi);
  statuscublas:= cublasAlloc(N, sizeof(double), @vk);
  statuscublas:= cublasAlloc(N, sizeof(double), @w);
  statuscublas:= cublasAlloc(N, sizeof(double), @xx);

  statuscublas:= cublasSetVector(Flength, sizeof(double),@Self.Values[0],1,@SelfValues[0],1);
  statuscublas:= cublasSetVector(N, sizeof(double),@B.Values[0],1,@r[0],1);
  statuscublas:= cublasSetVector(N, sizeof(double),@X.Values[0],1,@xx[0],1);

  bnrm2:= cublasDnrm2(N, r[0], 1);

  //* r = r - A*x; r is now residual */
  cublasDgemv('n', N, N, -1.0, SelfValues[0], N, xx[0], 1, 1.0, r[0], 1);
  rnorm2:= cublasDnrm2 (N, r[0], 1);
  error :=  rnorm2 / bnrm2;                 //* norm_r = norm(b) */
  if ( error < TOL ) then goto finish;     //* x is close enough already */


  for iter:= 1 to MAX_IT do
  begin
    cublasDgemv('n', N, N, -1.0, SelfValues[0], N, xx[0], 1, 1.0, r[0], 1);
    rnorm2:= cublasDnrm2(N,r[0],1);
    cublasDscal(N,1/rnorm2,r[0],1);
    cublasDcopy(N,r[0],1,V[0],1);
    s[0]:= rnorm2;

    for i:=0 to m-1 do
    begin
      cublasDcopy(N,V[i*N],1,vi[0],1);
      cublasDgemv('n', N, N, 1.0, SelfValues[0], N, vi[0], 1, 0.0, w[0], 1);
      for k:= 0 to i do
      begin
        cublasDcopy(N,V[k*N],1,vk[0],1);
        H[IDX2(k, i, m+1)]:= cublasDdot(N,w[0],1,vk[0],1);
        cublasDscal(N,-H[IDX2(k, i, m+1)],vk[0],1);
        cublasDaxpy(N,1.0,vk[0],1,w[0],1);
      end;
      H[IDX2(i+1, i, m+1)]:= cublasDnrm2(N,w[0],1);
      cublasDscal(N,1/H[IDX2(i+1, i, m+1)],w[0],1);
      cublasDcopy(N,w[0],1,V[(i+1)*N],1);
      for k:=0 to i-1 do
      begin
      kp1:= k+1;
      temp:= cs[k]*H[IDX2(k, i, m+1)] + sn[k]*H[IDX2(kp1, i, m+1)];
      H[IDX2(kp1, i, m+1)]:= -sn[k]*H[IDX2(k, i, m+1)] + cs[k]*H[IDX2(kp1, i, m+1)];
      H[IDX2(k, i, m+1)]:= temp;
      end;
      rotmat(H[IDX2(i, i, m+1)], H[IDX2(i+1, i, m+1)], sni, csi);
      sn[i]:= sni;
      cs[i]:= csi;
      temp:= cs[i]*s[i];
      s[i+1]:= -sn[i]*s[i];
      s[i]:= temp;
      H[IDX2(i, i, m+1)]:= cs[i]*H[IDX2(i, i, m+1)] + sn[i]*H[IDX2(i+1, i, m+1)];
      H[IDX2(i+1, i, m+1)]:= 0;
      error:= abs(s[i+1])/bnrm2;
      if ( error < TOL ) then
      begin
      setlength(s,(i));
      setlength(hhh,(i)*(i));
      for col:= 0 to i-1 do
        for row:= 0 to i-1 do
           hhh[IDX2(row, col, i)]:=h[IDX2(row, col, m+1)];

      statuscublas:= cublasAlloc(i, sizeof(double), @y);
      statuscublas:= cublasAlloc((i)*(i), sizeof(double), @hh);

      statuscublas:= cublasSetVector(i, sizeof(double),@s[0],1,@y[0],1);
      statuscublas:= cublasSetVector((i)*(i), sizeof(double),@HHH[0],1,@hh[0],1);

      cublasDtrsv(' ',' ',' ',(i), hh[0], (i), y[0], 1);
      cuda_error:= cublasGetError;
      cublasDgemv('N', N, (i), 1.0, V[0], N, y[0], 1, 1.0, xx[0] , 1);
      cuda_error:= cublasGetError;
      goto finish;
      end;
    end;
    setlength(s,m);
    setlength(hhh,m*m);
    for col:= 0 to m-1 do
      for row:= 0 to m-1 do
         hhh[IDX2(row, col, m)]:=h[IDX2(row, col, m+1)];
    statuscublas:= cublasAlloc(m, sizeof(double), @y);
    statuscublas:= cublasAlloc(m*m, sizeof(double), @hh);

    statuscublas:= cublasSetVector(m, sizeof(double),@s[0],1,@y[0],1);
    statuscublas:= cublasSetVector(m*m, sizeof(double),@H[0],1,@hh[0],1);

    cublasDtrsv(' ',' ',' ',m, hh[0], m, y[0], 1);
    cuda_error:= cublasGetError;
    cublasDgemv('N', N, m, 1.0, V[0], N, y[0], 1, 1.0, xx[0] , 1);
    cuda_error:= cublasGetError;
  end; // iter

  finish:
  statuscublas:= cublasGetVector(N, sizeof(double),@xx[0],1,@X.Values[0],1);

  statuscublas:= cublasFree(@SelfValues[0]);
  statuscublas:= cublasFree(@r[0]);
  statuscublas:= cublasFree(@V[0]);
  statuscublas:= cublasFree(@vi[0]);
  statuscublas:= cublasFree(@vk[0]);
  statuscublas:= cublasFree(@w[0]);
  statuscublas:= cublasFree(@xx[0]);

  result:= 0;
end;


end.
